// CHANGED 2004-07-28 Lutz Rogowski
#include <stdio.h>
#include <stdlib.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <sys/types.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>

#include <qapplication.h>

#include "dspoutoss.h" 

DspOutOss::DspOutOss( const QString &fileName, int fs )
	: DspOut()
{
	filename = fileName;
	setDeviceName( "oss" );
	audio_fd = -1;
	audio_stereo = false;
    fragsize = fs;
}

DspOutOss::~DspOutOss( void )
{
	printf("DspOutOss: ~\n");
	close( audio_fd );
	printf("DspOutOss end: ~\n");
}

#define FRAGSIZE_MASK 9
bool DspOutOss::openDevice( DeviceMode mode )
{
	int oflag;
	if( mode == ReadOnly ) {
		oflag = O_RDONLY;
	} else if( mode == WriteOnly ) {
		oflag = O_WRONLY;
	} else {
		oflag = O_RDWR;
	}
	if( devstate == DeviceOpened ) {
		lasterror = "Device Already Open";
		printf( "ERROR: %s\n", lasterror.ascii() );
		return false;
	}

	// Open immediately or else
	audio_fd = open( filename.ascii(), oflag | O_NONBLOCK );
	if( audio_fd == -1 ) {
		lasterror = "Open Failed";
		printf( "ERROR: %s\n", lasterror.ascii() );
		return false;
	}
   
	// Remove O_NONBLOCK
	int flags = fcntl( audio_fd, F_GETFL );
	flags &= ~O_NONBLOCK;
	fcntl( audio_fd, F_SETFL, flags );
    
	int format = AFMT_S16_LE;
	if( ioctl( audio_fd, SNDCTL_DSP_SETFMT, &format ) == -1 ) {
		lasterror = QString( "SETFMT " ) + QString( strerror( errno ) );
		printf( "ERROR: %s\n", lasterror.ascii() );
		return false;
	}

	if( format != AFMT_S16_LE ) {
		lasterror = "Format not supported";
		printf( "ERROR: %s\n", lasterror.ascii() );
		return false;
	}

	// keep fragsize less than 20ms !!
	//fragsize = 9;
	int frag = ( ( 32767 << 16 ) | fragsize ); // 7 orig val: = 2^7 = 128
	if( ioctl( audio_fd, SNDCTL_DSP_SETFRAGMENT, &frag ) ) {
		lasterror = QString( "SETFRAGMENT " ) + QString( strerror( errno ) );
		printf( "ERROR: %s\n", lasterror.ascii() );
		return false;
	}

	frag = 0;
	if( ioctl( audio_fd, SNDCTL_DSP_GETBLKSIZE, &frag ) ) {
		lasterror = QString( "GETBLKSIZE " ) + QString( strerror( errno ) );
		printf( "ERROR: %s\n", lasterror.ascii() );
		return false;
	}
	qDebug("FragSize %d (i.e %d samples) requested.\nFS is set to: %d ", fragsize,( 1<< fragsize  ), frag);

	int channels = 1;
	qDebug("Setting Channels to 1 ");
	if( ioctl( audio_fd, SNDCTL_DSP_CHANNELS, &channels ) == -1 ) {
		lasterror = QString( "DSP_CHANNELS " ) + QString( strerror( errno ) );
		printf( "ERROR: %s\n", lasterror.ascii() );
		return false;
	}

	if( channels != 1 && channels != 2 ) {
		lasterror = "Unsupported Number of Channels";
		printf( "ERROR: %s\n", lasterror.ascii() );
		return false;
	}

	if( channels == 2 ) {
		printf( "WARNING: nonmonophonic driver, forced into stereo mode (experimental)\n" );
		audio_stereo=true;
	} else {
		audio_stereo=false;
	}

	rate = 8000;
	qDebug("Setting rate to 8000 ");

	if( ioctl( audio_fd, SNDCTL_DSP_SPEED, &rate ) == -1 ) {
		lasterror = QString( "DSP_SPEED " ) + QString( strerror( errno ) );
		printf( "ERROR: %s\n", lasterror.ascii() );
		return false;
	}

	if( rate != 8000 ) {
		printf( "WARNING: driver rounded %d Hz request to %d Hz, off by %f%%\n",
			8000, rate, 100*((rate-8000)/8000.0) );
	}


	audio_buf_info info;

	if( mode == WriteOnly ) {
		if( ioctl( audio_fd, SNDCTL_DSP_GETOSPACE, &info ) == -1 ) {
			lasterror = QString( "GETOSPACE " ) + QString( strerror( errno ) );
			printf( "ERROR: %s\n", lasterror.ascii() );
			return false;
		}
	} else {
		if( ioctl( audio_fd, SNDCTL_DSP_GETISPACE, &info ) == -1 ) {
			lasterror = QString( "GETISPACE " ) + QString( strerror( errno ) );
			printf( "ERROR: %s\n", lasterror.ascii() );
			return false;
		}
        qDebug("Trying to set MIC as input source... ");
        //int dev = SOUND_MASK_MIC;
        int dev = SOUND_MASK_LINE1;
        if( ioctl( audio_fd, SOUND_MIXER_WRITE_RECSRC, &dev ) == -1 ) {
			lasterror = QString( "SOUND_MIXER_WRITE_RECSRC " ) + QString( strerror( errno ) );
			printf( "%s (no problem on AV500)\n", lasterror.ascii() );
		}

        dev = 0;
        qDebug("The following settings are used: ");
        if( ioctl( audio_fd, SOUND_MIXER_READ_IGAIN, &dev ) == -1 ) {
			lasterror = QString( "SOUND_MIXER_READ_IGAIN " ) + QString( strerror( errno ) );
			printf( "%s\n", lasterror.ascii() );
	} else {
			printf( "MIC INPUT GAIN %d%%  %d%% \n", dev & 0xff, (dev &0xff00) >> 8 );
        }

        dev = (90<<8) | 90;
	if( ioctl( audio_fd, SOUND_MIXER_WRITE_RECLEV, &dev ) == -1 ) {
			lasterror = QString( "SOUND_MIXER_WRITE_RECLEV " ) + QString( strerror( errno ) );
			printf( "%s\n", lasterror.ascii() );
	} 
	
        if( ioctl( audio_fd, SOUND_MIXER_READ_RECLEV, &dev ) == -1 ) {
			lasterror = QString( "SOUND_MIXER_READ_RECLEV " ) + QString( strerror( errno ) );
			printf( "%s\n", lasterror.ascii() );
		} else {
			printf( "MIC RECORD LEVEL %d%%  %d%% \n", dev & 0xff, (dev &0xff00) >> 8 );
        }

        dev = 0;
        if( ioctl( audio_fd, SOUND_MIXER_READ_MIC, &dev ) == -1 ) {
			lasterror = QString( "SOUND_MIXER_READ_MIC " ) + QString( strerror( errno ) );
			printf( "%s\n", lasterror.ascii() );
		} else {
			printf( "AMPLIFIER MIC %d%%  %d%% \n", dev & 0xff, (dev &0xff00) >> 8 );
        }

	}
	audio_buf.resize( info.fragsize * sizeof( short ) );
	qDebug("Resize buffer to %d ",info.fragsize * sizeof( short ) );

	lasterror = QString::null;
	devstate = DeviceOpened;

	qDebug("Open device finished :-)");
	return true;

}

bool DspOutOss::openDevice( int exist_fs )
{
	audio_fd = exist_fs;
	if( audio_fd == -1 ) {
		lasterror = "Open Failed";
		printf( "ERROR: %s\n", lasterror.ascii() );
		return false;
	}

	audio_buf_info info;
	if( ioctl( audio_fd, SNDCTL_DSP_GETISPACE, &info ) == -1 ) {
		lasterror = QString( "GETISPACE" ) + QString( strerror( errno ) );
		printf( "ERROR: %s\n", lasterror.ascii() );
		return false;
	}
	audio_buf.resize( info.fragsize * sizeof( short ) );

	lasterror = QString::null;
	devstate = DeviceOpened;
	return true;
}



// count is in 16-bit words in the smaller of the two buffers
void mono_to_stereo(void *b_from, void *b_to, size_t count)
{
    qDebug("ERROR mono_to_stereo ");
	int8_t *b1 = reinterpret_cast<int8_t*>(b_from);
	int8_t *b2 = reinterpret_cast<int8_t*>(b_to);
	for( size_t i=0; i < count; i++ ) {
		int8_t x0 = b1[2*i];
		int8_t x1 = b1[2*i+1];
		b2[4*i]   = x0;
		b2[4*i+1] = x1;
		b2[4*i+2] = x0;
		b2[4*i+3] = x1;
	}
}

// count is in 16-bit words in the smaller of the two buffers
void stereo_to_mono(void *b_from, void *b_to, size_t count)
{
    qDebug("ERROR stereo_to_mono ");
	int8_t *b1 = reinterpret_cast<int8_t*>(b_from);
	int8_t *b2 = reinterpret_cast<int8_t*>(b_to);
	for( size_t i=0; i < count; i++ ) {
		b2[2*i]   = b1[4*i];
		b2[2*i+1] = b1[4*i+1];
	}
}

bool DspOutOss::writeBuffer( void )
{
  
	if( devstate != DeviceOpened ) {
		lasterror = "Device Not Open";
		return false;
	}
	size_t count = audio_buf.getSize();
	void *buf;
	if( audio_stereo ) {
		buf = alloca(2*count);
		mono_to_stereo(audio_buf.getData(), buf, count/2);
		count *= 2;
	} else {
		buf = audio_buf.getData();
	}
	audio_buf_info info;
  
	if( ioctl( audio_fd, SNDCTL_DSP_GETOSPACE, &info ) == 0 ) {
		if( info.fragstotal - info.fragments > 15 ) { // 15 orig val
			// drop the fragment if the buffer starts to fill up
            //qDebug("drop ");
			return true;
		}
	}
   
    int br;
    if( (br = write( audio_fd, audio_buf.getData(), audio_buf.getSize() )) != audio_buf.getSize()  )
	  qDebug("DspOutOss::writeBuffer ERROR: buf size: %d - written %d ",audio_buf.getSize(), br );
   
	return true;
}

unsigned int DspOutOss::readableBytes( void )
{
	audio_buf_info info;
    // if (QApplication::desktop()->width() > 320 ) 
    {
        struct timeval timeout;
        fd_set read_fds;
        
        if( devstate != DeviceOpened ) {
            return 0;
        }
#if 0
        timeout.tv_sec = 0;
        timeout.tv_usec = 0;
        FD_ZERO( &read_fds );
        FD_SET( audio_fd, &read_fds );

        if( select( audio_fd + 1, &read_fds, NULL, NULL, &timeout ) == -1 ) {
            return 0;
        }
        if( !FD_ISSET( audio_fd, &read_fds ) ) {
            return 0;
        }
#endif
    }
	if( ioctl( audio_fd, SNDCTL_DSP_GETISPACE, &info ) == -1 ) {
		lasterror = strerror( errno );
		return 0;
	}
    // qDebug("r: %d ", info.bytes);
	return info.bytes;
}


int DspOutOss::readBuffer( int bytes )
{
   
	if( devstate != DeviceOpened ) {
		lasterror = "Device Not Open";
		return 0;
	}

	audio_buf.resize( bytes );
	size_t count = bytes;

	void *buf;
	if( audio_stereo ) {
		count *= 2;
		buf = alloca(count);
	} else {
		buf = audio_buf.getData();
	}
	int rc = read( audio_fd, buf, count );
	if( rc != count ) {
		printf("warning: asked microphone for %d got %d\n", count, rc);
	}
				
	if( audio_stereo ) {
		stereo_to_mono(buf, audio_buf.getData(), count/4);
	}
    if ( rc > 0 ) return rc;
	return 0;
}

